#!/usr/bin/env python

"""
Acorn Electron ULA simulation.
"""

from array import array
from itertools import repeat

LINES_PER_ROW = 8       # the number of pixel lines per character row
MAX_HEIGHT = 256        # the height of the screen in pixels
MAX_SCANLINE = 312      # the number of scanlines in each frame
MAX_WIDTH = 640         # the width of the screen in pixels
MAX_SCANPOS = 1024      # the number of positions in each scanline
SCREEN_LIMIT = 0x8000   # the first address after the screen memory
MAX_MEMORY = 0x10000    # the number of addressable memory locations
BLANK = (0, 0, 0)

def update(ula):

    """
    Update the 'ula' for one frame. Return the resulting screen.
    """

    video = ula.video

    i = 0
    limit = MAX_SCANLINE * MAX_SCANPOS
    while i < limit:
        ula.update()
        video.update()
        i += 1
    return video.screen

class Video:

    """
    A class representing the video circuitry.
    """

    def __init__(self):
        self.screen = array("B", repeat(0, MAX_WIDTH * 3 * MAX_HEIGHT))
        self.colour = BLANK
        self.csync = 1
        self.hs = 1
        self.reset()

    def reset(self):
        self.pos = 0

    def update(self):
        if self.csync:
            if self.hs:
                self.screen[self.pos] = self.colour[0]; self.pos += 1
                self.screen[self.pos] = self.colour[1]; self.pos += 1
                self.screen[self.pos] = self.colour[2]; self.pos += 1
        else:
            self.pos = 0

class ULA:

    """
    A class providing the ULA functionality. Instances of this class refer to
    the system memory, maintain internal state (such as information about the
    current screen mode), and provide outputs (such as the current pixel
    colour).
    """

    modes = [
        (640, 1, 32), (320, 2, 32), (160, 4, 32),   # (width, depth, rows)
        (640, 1, 25), (320, 1, 32), (160, 2, 32),
        (320, 1, 25)
        ]

    palette = range(0, 8) * 2

    def __init__(self, memory, video):

        "Initialise the ULA with the given 'memory' and 'video'."

        self.memory = memory
        self.video = video
        self.set_mode(6)

        # Internal state.

        self.buffer = [0] * 8

        self.reset()

    def reset(self):

        "Reset the ULA."

        self.vsync()

    def set_mode(self, mode):

        """
        For the given 'mode', initialise the...

          * width in pixels
          * colour depth in bits per pixel
          * number of character rows
          * character row size in bytes
          * screen size in bytes
          * default screen start address
          * horizontal pixel scaling factor
          * line spacing in pixels
          * number of entries in the pixel buffer

        The ULA should be reset after a mode switch in order to cleanly display
        a full screen.
        """

        self.width, self.depth, rows = self.modes[mode]

        columns = (self.width * self.depth) / 8                             # bits read -> bytes read
        row_size = columns * LINES_PER_ROW

        # Memory access configuration.
        # Note the limitation on positioning the screen start.

        screen_size = row_size * rows
        self.screen_start = (SCREEN_LIMIT - screen_size) & 0xff00
        self.screen_size = SCREEN_LIMIT - self.screen_start

        # Scanline configuration.

        self.xscale = MAX_WIDTH / self.width                                # pixel width in display pixels
        self.spacing = MAX_HEIGHT / rows - LINES_PER_ROW                    # pixels between rows

        # Start of unused region.

        self.footer = rows * LINES_PER_ROW
        self.margin = MAX_SCANLINE - rows * (LINES_PER_ROW + self.spacing) + self.spacing

        # Internal pixel buffer size.

        self.buffer_limit = 8 / self.depth

    def vsync(self):

        "Signal the start of a frame."

        self.line_start = self.address = self.screen_start
        self.line = self.line_start % LINES_PER_ROW
        self.ssub = 0
        self.y = 0
        self.reset_horizontal()

        # Signal the video circuit.

        self.csync = self.video.csync = 1

    def hsync(self):

        "Signal the end of a scanline."

        self.y += 1
        self.reset_horizontal()

        # Support spacing between character rows.

        if self.ssub:
            self.ssub -= 1
            return

        self.line += 1

        # If not on a row boundary, move to the next line.

        if self.line % LINES_PER_ROW:
            self.address = self.line_start + 1
            self.wrap_address()

        # After the end of the last line in a row, the address should already
        # have been positioned on the last line of the next column.

        else:
            self.address -= LINES_PER_ROW - 1
            self.wrap_address()

            # Test for the footer region.

            if self.spacing and self.line == self.footer:
                self.ssub = self.margin
                return

            # Support spacing between character rows.

            self.ssub = self.spacing

        self.line_start = self.address

    def reset_horizontal(self):

        "Reset horizontal state."

        self.x = 0
        self.buffer_index = self.buffer_limit # need refill

        # Signal the video circuit.

        self.hs = self.video.hs = 1

    def update(self):

        """
        Update the pixel colour by reading from the pixel buffer.
        """

        # Detect the end of the line.

        if self.x >= MAX_WIDTH:
            if self.x == MAX_WIDTH:
                self.hs = self.video.hs = 0

            # Detect the end of the scanline.

            elif self.x == MAX_SCANPOS:
                self.hsync()

                # Detect the end of the frame.

                if self.y == MAX_SCANLINE:
                    self.vsync()

                # Detect the end of the screen.

                elif self.y == MAX_HEIGHT:
                    self.csync = self.video.csync = 0

        # Detect spacing between character rows.

        if self.ssub:
            self.video.colour = BLANK

        # Detect horizontal and vertical sync conditions.

        elif not self.hs or not self.csync:
            pass

        # For pixels within the frame, obtain and output the value.

        else:

            # Scale pixels horizontally, only accessing the next pixel value
            # after the required number of scan positions.

            if self.x % self.xscale == 0:
                self.buffer_index += 1

            # Fill the buffer once all values have been read.

            if self.buffer_index >= self.buffer_limit:
                self.buffer_index = 0
                self.fill_pixel_buffer()

            self.video.colour = self.buffer[self.buffer_index]

        self.x += 1

    def fill_pixel_buffer(self):

        """
        Fill the pixel buffer by translating memory content for the current
        mode.
        """

        byte_value = self.memory[self.address]

        i = 0
        for colour in decode(byte_value, self.depth):
            self.buffer[i] = get_physical_colour(self.palette[colour])
            i += 1

        # Advance to the next column.

        self.address += LINES_PER_ROW
        self.wrap_address()

    def wrap_address(self):
        if self.address >= SCREEN_LIMIT:
            self.address -= self.screen_size

    # Convenience methods.

    def fill(self, start, end, value):
        fill(self.memory, start, end, value)

def get_physical_colour(value):

    """
    Return the physical colour as an RGB triple for the given 'value'.
    """

    return value & 1, value >> 1 & 1, value >> 2 & 1

def decode(value, depth):

    """
    Decode the given byte 'value' according to the 'depth' in bits per pixel,
    returning a sequence of pixel values.
    """

    if depth == 1:
        return (value >> 7, value >> 6 & 1, value >> 5 & 1, value >> 4 & 1,
                value >> 3 & 1, value >> 2 & 1, value >> 1 & 1, value & 1)
    elif depth == 2:
        return (value >> 6 & 2 | value >> 3 & 1, value >> 5 & 2 | value >> 2 & 1,
                value >> 4 & 2 | value >> 1 & 1, value >> 3 & 2 | value & 1)
    elif depth == 4:
        return (value >> 4 & 8 | value >> 3 & 4 | value >> 2 & 2 | value >> 1 & 1,
                value >> 3 & 8 | value >> 2 & 4 | value >> 1 & 2 | value & 1)
    else:
        raise ValueError, "Only depths of 1, 2 and 4 are supported, not %d." % depth

# Convenience functions.

def encode(values, depth):

    """
    Encode the given 'values' according to the 'depth' in bits per pixel,
    returning a byte value for the pixels.
    """

    result = 0

    if depth == 1:
        for value in values:
            result = result << 1 | (value & 1)
    elif depth == 2:
        for value in values:
            result = result << 1 | (value & 2) << 3 | (value & 1)
    elif depth == 4:
        for value in values:
            result = result << 1 | (value & 8) << 3 | (value & 4) << 2 | (value & 2) << 1 | (value & 1)
    else:
        raise ValueError, "Only depths of 1, 2 and 4 are supported, not %d." % depth

    return result

def get_ula():

    "Return a ULA initialised with a memory array and video."

    return ULA(get_memory(), get_video())

def get_video():

    "Return a video circuit."

    return Video()

def get_memory():

    "Return an array representing the computer's memory."

    return [0] * MAX_MEMORY

def fill(memory, start, end, value):
    for i in xrange(start, end):
        memory[i] = value

# Test program providing coverage (necessary for compilers like Shedskin).

if __name__ == "__main__":
    ula = get_ula()
    ula.set_mode(2)
    ula.fill(0x5800 - 320, 0x8000, encode((2, 7), 4))

    # Make a simple two-dimensional array of tuples (three-dimensional in pygame
    # terminology).

    a = update(ula)

# vim: tabstop=4 expandtab shiftwidth=4
